/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Clipboard
Version:   1.0
Status:    Stable
ID:        ID_CLIPBOARD
Category:  I/O
Date:      May 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2001-2004.  All rights reserved.
Short:     The Clipboard class manages cut, copy and paste operations.
Keywords:  clipboard, cut, copy, paste, text, image

###DESCRIPTION###
<p>The Clipboard class manages data transfer between applications on behalf
of the user.  It holds a data cache of clipped items that originate from
source applications, and these can be retrieved and 'pasted' into target
applications.  The Clipboard class is provided for the primary purpose of
allowing applications to implement the traditional 'cut', 'copy' and
'paste' actions.</p>

<p>There is no limit on the amount of clipboard objects allowed in the system,
but they all control the same set of data for the logged-in user.  Items that are
passed to the clipboard object are stored in the user directory
"temp:clipboard/".  Clipped items are persistent, so if the machine is turned
off, all currently clipped items will still be accessible on the next restart.
There is a limit on the amount of clipped items that can be stored in the
clipboard object.  Only 1 of each datatype is permitted (for example, only one 'image'
clip may exist at any time) and there is a preset limitation on the total
number of clips that can be stored in the history cache.</p>

###END###

*****************************************************************************/

#include <pandora/data/config.h>
#include <pandora/files/file.h>
#include <pandora/input/clipboard.h>
#include <pandora/graphics/text.h>
#include <pandora/graphics/window.h>
#include <pandora/system/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_CLIPBOARD, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2001-2004.  All rights reserved.",
   "May 2004",
   "Clipboard"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR ClipboardClass = 0;
static struct StringsBase *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR modStrings = 0, modFileSystem = 0;

struct FieldDef glDatatypes[] = {
   { "data",   CLIP_DATA },
   { "audio",  CLIP_AUDIO },
   { "image",  CLIP_IMAGE },
   { "file",   CLIP_FILE },
   { "object", CLIP_OBJECT },
   { "text",   CLIP_TEXT },
   { NULL, NULL }
};

/*****************************************************************************
** Class definition
*/

#define MAX_CLIPS 10     /* Maximum number of clips stored in the historical buffer */

struct ClipEntry {
   LONG     Datatype;    /* The type of data clipped */
   LONG     Flags;       /* CLF_DELETE may be set for the 'cut' operation */
   CLASSID  ClassID;     /* Class ID that is capable of managing the clip data, if it originated from an object */
   MEMORYID Files;       /* List of file locations, separated with semi-colons, referencing all the data in this clip entry */
   LONG     FilesLen;    /* Complete byte-length of the Files string */
   UWORD    ID;          /* Unique identifier for the clipboard entry */
   WORD     TotalItems;  /* Total number of items in the clip-set */
};

struct Clipboard {
   OBJECT_HEADER
   LONG Response;   /* Indicates the user's reponse to PasteFiles() */
   LONG Static;

   PRIVATE_FIELDS
   LONG Counter;
   UBYTE CrashProtect;
   OBJECTPTR ProgressText;
   LARGE ProgressTime;
   OBJECTID ProgressTarget;
   OBJECTID ProgressWindow;
};

enum {
   RSP_CANCEL=0,
   RSP_YES,
   RSP_YESALL,
   RSP_NO
};

static struct FieldDef ResponseLookup[] = {
   { "CANCEL", RSP_CANCEL },
   { "YES",    RSP_YES },
   { "YESALL", RSP_YESALL },
   { "NO",     RSP_NO },
   { "NOALL",  RSP_CANCEL },
   { NULL, NULL }
};

static struct FieldArray ClipboardFields[] = {
   { "Response", 0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ResponseLookup, NULL, NULL }, /* NB: Is writeable so that our confirmation boxes can change it */
   { "Static",   0, FDF_LONG|FDF_RI,            0, NULL, NULL },
   END_FIELD
};

/*****************************************************************************
** Clipboard actions
*/

static ERROR CLIPBOARD_AccessObject(struct Clipboard *, APTR);
static ERROR CLIPBOARD_ActionNotify(struct Clipboard *, struct acActionNotify *);
static ERROR CLIPBOARD_Clear(struct Clipboard *, APTR);
static ERROR CLIPBOARD_ClosingTag(struct Clipboard *, APTR);
static ERROR CLIPBOARD_DataChannel(struct Clipboard *, struct acDataChannel *);
static ERROR CLIPBOARD_Free(struct Clipboard *, APTR);
static ERROR CLIPBOARD_GetUnlistedField(struct Clipboard *, struct acGetUnlistedField *);
static ERROR CLIPBOARD_Init(struct Clipboard *, APTR);
static ERROR CLIPBOARD_ReleaseObject(struct Clipboard *, APTR);

static struct ActionArray ClipboardActions[] = {
   { AC_AccessObject,     CLIPBOARD_AccessObject },
   { AC_ActionNotify,     CLIPBOARD_ActionNotify },
   { AC_Clear,            CLIPBOARD_Clear },
   { AC_ClosingTag,       CLIPBOARD_ClosingTag },
   { AC_DataChannel,      CLIPBOARD_DataChannel },
   { AC_Free,             CLIPBOARD_Free },
   { AC_GetUnlistedField, CLIPBOARD_GetUnlistedField },
   { AC_Init,             CLIPBOARD_Init },
   { AC_ReleaseObject,    CLIPBOARD_ReleaseObject },
   { NULL, NULL }
};

/*****************************************************************************
** Clipboard methods
*/

static ERROR CLIPBOARD_AddClipboardObject(struct Clipboard *, struct mtAddClipboardObject *);
static ERROR CLIPBOARD_AddClipboardObjects(struct Clipboard *, struct mtAddClipboardObjects *);
static ERROR CLIPBOARD_AddClipboardFiles(struct Clipboard *, struct mtAddClipboardFiles *);
static ERROR CLIPBOARD_AddClipboardText(struct Clipboard *, struct mtAddClipboardText *);
static ERROR CLIPBOARD_DeleteClip(struct Clipboard *, struct mtDeleteClip *);
static ERROR CLIPBOARD_GetClipboardFiles(struct Clipboard *, struct mtGetClipboardFiles *);
static ERROR CLIPBOARD_PasteClipboardFiles(struct Clipboard *, struct mtPasteClipboardFiles *);

static struct FunctionField argsAddClipboardFiles[]   = { { "Datatype", ARG_LONG }, { "Files",    ARG_STR  }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsAddClipboardObject[]  = { { "Datatype", ARG_LONG }, { "Object",   ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsAddClipboardObjects[] = { { "Datatype", ARG_LONG }, { "Objects",  ARG_LONG }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsAddClipboardText[]    = { { "String",   ARG_STR  }, { NULL, NULL } };
static struct FunctionField argsDeleteClip[]          = { { "Datatype", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsGetClipboardFiles[]   = { { "Datatype", ARG_LONG }, { "Index", ARG_LONG }, { "Files",  ARG_STRING }, { "Flags", ARG_LONG }, { NULL, NULL } };
static struct FunctionField argsPasteClipboardFiles[] = { { "Destination", ARG_STR }, { "Target", ARG_LONG }, { NULL, NULL } };

static struct MethodArray ClipboardMethods[] = {
   { MT_AddClipboardFiles,   CLIPBOARD_AddClipboardFiles,   "AddFiles",   argsAddClipboardFiles,   sizeof(struct mtAddClipboardFiles) },
   { MT_AddClipboardObject,  CLIPBOARD_AddClipboardObject,  "AddObject",  argsAddClipboardObject,  sizeof(struct mtAddClipboardObject) },
   { MT_AddClipboardObjects, CLIPBOARD_AddClipboardObjects, "AddObjects", argsAddClipboardObjects, sizeof(struct mtAddClipboardObjects) },
   { MT_AddClipboardText,    CLIPBOARD_AddClipboardText,    "AddText",    argsAddClipboardText,    sizeof(struct mtAddClipboardText) },
   { MT_GetClipboardFiles,   CLIPBOARD_GetClipboardFiles,   "GetFiles",   argsGetClipboardFiles,   sizeof(struct mtGetClipboardFiles) },
   { MT_PasteClipboardFiles, CLIPBOARD_PasteClipboardFiles, "PasteFiles", argsPasteClipboardFiles, sizeof(struct mtPasteClipboardFiles) },
   { MT_DeleteClip,          CLIPBOARD_DeleteClip,          "DeleteClip", argsDeleteClip,          sizeof(struct mtDeleteClip) },
   { NULL, NULL, NULL, NULL, NULL }
};

/****************************************************************************/

static void AddClip(struct Clipboard *Self, LONG Datatype, STRING Files, LONG Flags, CLASSID ClassID, LONG TotalItems);
static void FreeClip(struct Clipboard *, struct ClipEntry *);
static void RunScript(STRING, OBJECTID, STRING);
static LONG PasteFeedback(struct FileFeedback *);

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   MEMORYID memoryid;

   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   memoryid = RPM_Clipboard;
   AllocMemory(MAX_CLIPS * sizeof(struct ClipEntry), MEM_UNTRACKED|MEM_PUBLIC|MEM_RESERVED|MEM_NOBLOCKING, NULL, &memoryid);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ClipboardClass, NULL,
      FID_BaseClassID|TLONG, ID_CLIPBOARD,
      FID_Version|TFLOAT,    VER_CLIPBOARD,
      FID_Name|TSTRING,      "Clipboard",
      FID_Category|TLONG,    CCF_IO,
      FID_Actions|TPTR,      ClipboardActions,
      FID_Methods|TPTR,      ClipboardMethods,
      FID_Fields|TPTR,       ClipboardFields,
      FID_Size|TLONG,        sizeof(struct Clipboard),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (modStrings)     { acFree(modStrings);     modStrings  = NULL; }
   if (modFileSystem)  { acFree(modFileSystem);  modFileSystem = NULL; }
   if (ClipboardClass) { acFree(ClipboardClass); ClipboardClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: GetDatatype()
*/

static STRING GetDatatype(LONG Datatype)
{
   WORD i;

   for (i=0; glDatatypes[i].Name; i++) {
      if (Datatype IS glDatatypes[i].Value) return(glDatatypes[i].Name);
   }

   return("unknown");
}

/*****************************************************************************
** Action: AccessObject()
*/

static ERROR CLIPBOARD_AccessObject(struct Clipboard *Self, APTR Void)
{
   if (Self->CrashProtect) {
      /* If CrashProtect is set, the clipboard was never released
      ** correctly.  Potentially the clipboard could be damaged, so we
      ** have to reset it to be safe.
      */

      DPrintF("@AccessObject():","[Clipboard] Clipboard has crashed earlier - cleaning up now.");
#if 0
      WORD i;

      for (i=0; i < ARRAYSIZE(Self->Clips); i++) {
         if (Self->Clips[i].Files) {
            MEMORYID memid;
            memid = Self->Clips[i].Files;
            Self->Clips[i].Files = NULL;
            FreeMemoryID(memid);
         }
      }
      ClearMemory(Self->Clips, sizeof(Self->Clips));
      Self->Counter = 0;
#endif
   }

   Self->CrashProtect = 1; /* Turn on crash protection */
   return(ERR_Okay);
}

/*****************************************************************************
** Clipboard: ActionNotify
*/

static ERROR CLIPBOARD_ActionNotify(struct Clipboard *Self, struct acActionNotify *Args)
{
   if ((Args->ActionID IS MT_CloseWindow) AND (Args->ObjectID IS Self->ProgressWindow)) {
      /* The user has closed the progress window.  The copy process will
      ** continue as normal, but we won't keep bothering the user with our progress.
      */
      acFreeID(Self->ProgressWindow);
      Self->ProgressWindow = NULL;
      Self->ProgressText   = NULL;
      Self->ProgressTarget = -1; /* Set to -1 to stop the window from popping up again */
   }
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  AddClipboardFiles
Short: Add files to the clipboard.

###DESCRIPTION###
<p>This method is used to add file(s) of information to the clipboard.  You add
files to the clipboard as a string of locations, each separated by a semi-colon,
for copying to the clipboard data cache.  The information may then be used by
other applications if they understand the clipboard content.  You are required
to specify the type of data that you are adding through the Datatype parameter.
Recognised data types are:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">CLIP_AUDIO</td><td>Audio clips are any type of audio that is recognised by the Sound class (such as wav's and mp3's).</td></tr>
<tr><td width="10%">CLIP_DATA</td><td>Raw information that is uncategorised can be defined as a data clip.</td></tr>
<tr><td width="10%">CLIP_FILE</td><td>File references are stored as file clips.  This type is typically used by file managers and file dialog boxes.</td></tr>
<tr><td width="10%">CLIP_IMAGE</td><td>Images that are recognised by the Picture class may be stored as this type (such as jpeg's and png files).</td></tr>
<tr><td width="10%">CLIP_TEXT</td><td>Standard ASCII and UTF-8 formatted text should be identified through this clip type.</td></tr>
</table>

<p>If you specify CLIP_FILE as the Datatype, the location strings will be copied
directly to the clipboard and the original file source will not be cached.
If the CLIP_FILE type is used in conjunction with the CLF_DELETE flag then the
original file source will be deleted if it is copied from the clipboard successfully
(this is typically used to implement the 'cut' operation).</p>

<p>Each time that a new set of clip data is added to the clipboard, the
history buffer is filled by a new clipping entry.  When the history buffer is full,
the oldest entry is removed each time that a new clip is added.</p>

<p>Optional flags that may be passed to this method are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flags</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td>CLF_DELETE</td><td>May be set in conjunction with the CLIP_FILE data type if the original file source should be deleted after use.</td></tr>
<tr><td>CLF_EXTEND</td><td>Instead of replacing the existing clipboard data, add the clip item to what already exists for the specified datatype.</td></tr>
</table>

###INPUT###
LONG/Datatype: Set this argument to indicate the type of data you are copying to the clipboard.
STRING/Files: The locations of the files to add.
LONG/Flags: Optional flags.

###RESULT###
ERR_Okay: The files were added to the clipboard.
ERR_Args: Invalid arguments were specified.
ERR_MissingLoation: The Files argument was not correctly specified.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_AddClipboardFiles(struct Clipboard *Self, struct mtAddClipboardFiles *Args)
{
   struct ClipEntry *clips;
   UBYTE buffer[256];
   OBJECTPTR file;
   STRING src;
   LONG len;
   WORD i, totalitems;

   if (!Args) return(ObjectError(Self, MT_AddClipboardFiles, ERR_Args));
   if ((!Args->Files) OR (!Args->Files[0])) return(ObjectError(Self, MT_AddClipboardFiles, ERR_MissingLocation));

   DPrintF("~AddFiles()","[Clipboard:%d] Files: %s", Self->Head.UniqueID, Args->Files);

   totalitems = 0;
   if (Args->Datatype IS CLIP_FILE) {
      /* When the file datatype is used, we will just store the file
      ** locations directly and not perform data manipulations or caching.
      */

      len = StrLength(Args->Files) + 1;

      {
         UBYTE files[len];

         StrCopy(Args->Files, files, len);

         /*** Count the total number of files specified in the list ***/

         totalitems = 1;
         src = Args->Files;
         while (*src) {
            if (*src IS ';') totalitems++;
            src++;
         }

         AddClip(Self, Args->Datatype, files, Args->Flags & (CLF_DELETE|CLF_EXTEND), NULL, totalitems);
      }
   }
   else {
      /* Scan the location string and copy the files to the clipboard cache.
      ** Each clipboard file will be given the location:
      **
      **    temp:clipboard/datatype.xxx
      **
      ** Where 'datatype' is the name of the datatype and 'xxx' is an integer counter.
      */

      src = Args->Files;
      totalitems = 0;

      if (Args->Flags & CLF_EXTEND) {
         /* If the extend option is used, start counting from the existing
         ** total number of items.
         */

         if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
            for (i=0; i < MAX_CLIPS; i++) {
               if (clips[i].Datatype IS Args->Datatype) {
                  totalitems = clips[i].TotalItems;
                  break;
               }
            }
            ReleaseMemory(clips);
         }
         else ObjectError(Self, MT_AddClipboardFiles, ERR_AccessMemory);
      }

      while (*src) {
         for (i=0; (*src) AND (*src != ';') AND (i < sizeof(buffer)-1); src++); buffer[i] = *src;
         if (*src IS ';') src++;
         buffer[i] = 0;

         /*** Open the source file ***/

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, buffer,
            FID_Flags|TLONG,      FL_READ,
            TAGEND) IS ERR_Okay) {

            /*** Copy it to the destination clipboard cache file ***/

            StrFormat(buffer, sizeof(buffer), "temp:clipboard/%s%d.%.3d", GetDatatype(Args->Datatype), Self->Counter + 1, totalitems++);

            ActionTags(MT_CopyFile, file, buffer);

            acFree(file);
         }
      }

      AddClip(Self, Args->Datatype, NULL, Args->Flags & (CLF_DELETE|CLF_EXTEND), NULL, totalitems);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  AddClipboardObject
Short: Extract data from an object and add it to the clipboard.

###DESCRIPTION###
<p>This method is a simple implementation of the AddClipboardObjects method
and is intended primarily for script usage.  Please see the AddClipboardObjects
method for details on adding objects to the clipboard.</p>

###INPUT###
LONG/Datatype: The type of data that you want the object data to be recognised as, or NULL for automatic recognition.
OBJECTID/ObjectID: The object containing the data to add.
LONG/Flags: Optional flags.

###RESULT###
ERR_Okay: The object was added to the clipboard.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_AddClipboardObject(struct Clipboard *Self, struct mtAddClipboardObject *Args)
{
   struct mtAddClipboardObjects add;
   OBJECTID objects[2];

   if (!Args) return(ObjectError(Self, MT_AddClipboardObject, ERR_Args));

   objects[0] = Args->ObjectID;
   objects[1] = NULL;
   add.Objects = &Args->ObjectID;
   add.Flags   = Args->Flags;
   return(CLIPBOARD_AddClipboardObjects(Self, &add));
}

/*****************************************************************************

###METHOD###
Name:  AddClipboardObjects
Short: Extract data from objects and add it all to the clipboard.

###DESCRIPTION###
<p>Data can be saved to the clipboard directly from an object if the object's
class supports the SaveToObject action.  The clipboard will ask that the
object save its data directly to a cache file, completely removing the need
for you to save the object data to an interim file for the clipboard.</p>

<p>Certain classes are recognised by the clipboard system and will be added to
the correct datatype automatically (for instance, Picture objects will be put
into the CLIP_IMAGE data category).  If an object's class is not recognised by
the clipboard system then the data will be stored in the CLIP_OBJECT category to
signify that there is a class in the system that recognises the data.  If you
want to over-ride any aspect of this behaviour, you need to force the Datatype
parameter with one of the available CLIP* types.</p>

<p>This method supports groups of objects in a single clip, thus requires you
to pass an array of object ID's, terminated with a NULL entry.</p>

<p>Optional flags that may be passed to this method are the same as those
specified in the AddClipboardFiles method.  The CLF_DELETE flag has no effect
on objects.</p>

<p>This method should always be called directly and not messaged to the
clipboard, unless you are able to guarantee that the source objects are shared.</p>

###INPUT###
LONG/Datatype: The type of data that you want the object data to be recognised as, or NULL for automatic recognition.
*OBJECTID/Objects: Array of shared object ID's to add to the clipboard.
LONG/Flags: Optional flags.

###RESULT###
ERR_Okay: The objects were added to the clipboard.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_AddClipboardObjects(struct Clipboard *Self, struct mtAddClipboardObjects *Args)
{
   OBJECTPTR object;
   OBJECTID *list;
   UBYTE location[100];
   LONG datatype;
   CLASSID classid;
   WORD i;

   if ((!Args) OR (!Args->Objects) OR (!Args->Objects[0])) return(ObjectError(Self, MT_AddClipboardObjects, ERR_Args));

   DPrintF("~AddObjects()","[Clipboard:%d]", Self->Head.UniqueID);

   /* Use the SaveToObject action to save each object's data to the
   ** clipboard storage area.  The class ID for each object is also
   ** recorded.
   */

   classid  = NULL;
   datatype = NULL;
   list = Args->Objects;
   for (i=0; list[i]; i++) {
      if (AccessObject(list[i], 5000, &object) IS ERR_Okay) {
         if (!classid) classid = object->ClassID;

         if (classid IS object->ClassID) {
            datatype = Args->Datatype;

            if (!datatype) {
               if (object->ClassID IS ID_PICTURE) {
               StrFormat(location, sizeof(location), "temp:clipboard/image%d.%.3d", Self->Counter + 1, i);
               datatype = CLIP_IMAGE;
               }
               else if (object->ClassID IS ID_SOUND) {
                  StrFormat(location, sizeof(location), "temp:clipboard/audio%d.%.3d", Self->Counter + 1, i);
                  datatype = CLIP_AUDIO;
               }
               else {
                  StrFormat(location, sizeof(location), "temp:clipboard/object%d.%.3d", Self->Counter + 1, i);
                  datatype = CLIP_OBJECT;
               }
            }
            else {
               /*** Use the specified datatype ***/
               StrFormat(location, sizeof(location), "temp:clipboard/%s.%.3d", GetDatatype(datatype), i);
            }

            SaveObjectToFile(object, location, NULL);
         }

         ReleaseObject(object);
      }
   }

   AddClip(Self, datatype, NULL, Args->Flags & CLF_EXTEND, NULL, i);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  AddClipboardText
Short: Adds a block of text to the clipboard.

###DESCRIPTION###
<p>Text can be added to the clipboard using the AddClipboardText method.  This
is the simplest way of passing text to the clipboard, although passing text
through the data channel system may also be convenient in certain circumstances.
Text is passed to the clipboard via the String parameter and it must be
terminated with a null byte.</p>

###INPUT###
STRING/String: The text to add to the clipboard.

###RESULT###
ERR_Okay: The object was added to the clipboard.
ERR_Args: Invalid arguments were specified.
ERR_File: Failed to open a new file for output.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_AddClipboardText(struct Clipboard *Self, struct mtAddClipboardText *Args)
{
   struct acWrite write;
   OBJECTPTR file;
   UBYTE buffer[200];

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   DPrintF("~AddText()","[Clipboard:%d]", Self->Head.UniqueID);

   StrFormat(buffer, sizeof(buffer), "temp:clipboard/text%d.000", Self->Counter + 1);

   if (CreateObject(ID_FILE, NULL, &file, NULL,
         FID_Location|TSTRING,  buffer,
         FID_Flags|TLONG,       FL_NEW|FL_WRITE,
         FID_Permissions|TLONG, PERMIT_READ|PERMIT_WRITE,
         TAGEND) IS ERR_Okay) {

      write.Buffer = Args->String;
      write.Length = StrLength(Args->String);
      Action(AC_Write, file, &write);

      AddClip(Self, CLIP_TEXT, NULL, NULL, NULL, 1);

      acFree(file);
      StepBack();
      return(ERR_Okay);
   }
   else return(StepObjectError(Self, MT_AddClipboardText, ERR_File));
}

/*****************************************************************************

###ACTION###
Name:  Clear
Short: Destroys all cached data that is stored in the clipboard.
###END###

*****************************************************************************/

static ERROR CLIPBOARD_Clear(struct Clipboard *Self, APTR Void)
{
   struct ClipEntry *clips;

   /*** Delete the clipboard directory and all content ***/

   DeleteFile("temp:clipboard/");
   MakeDirectory("temp:clipboard", PERMIT_READ|PERMIT_WRITE);

   /*** Annihilate all historical clip information ***/

   if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
      ClearMemory(&clips, MAX_CLIPS * sizeof(struct ClipEntry));
      ReleaseMemory(clips);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Clear, ERR_AccessMemory));
}

/*****************************************************************************
** Action: ClosingTag
*/

static ERROR CLIPBOARD_ClosingTag(struct Clipboard *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      acFree(Self);
   }
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: This action can be used to place data in a clipboard.

###DESCRIPTION###
<p>Data can be sent to a Clipboard object via the DataChannel action.
Currently, only the DATA_TEXT type is supported.  All data that is sent to a
Clipboard object through this action will replace any stored information that
matches the given data type.</p>

###END###

*****************************************************************************/

static ERROR CLIPBOARD_DataChannel(struct Clipboard *Self, struct acDataChannel *Args)
{
   struct acWrite write;
   OBJECTPTR file;
   UBYTE buffer[200];

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_TEXT) AND (Args->Version IS 1)) {
      DPrintF("DataChannel()","[Clipboard:%d] Copying text to the clipboard.", Self->Head.UniqueID);

      StrFormat(buffer, sizeof(buffer), "temp:clipboard/text%d.000", Self->Counter + 1);

      if (CreateObject(ID_FILE, NULL, &file, NULL,
            FID_Location|TSTRING,  buffer,
            FID_Flags|TLONG,       FL_NEW|FL_WRITE,
            FID_Permissions|TLONG, PERMIT_READ|PERMIT_WRITE,
            TAGEND) IS ERR_Okay) {

         write.Buffer = Args->Buffer;
         write.Length = Args->Size;
         if (Action(AC_Write, file, &write) != ERR_Okay) {
            acFree(file);
            return(ObjectError(Self, ERH_DataChannel, ERR_Write));
         }

         AddClip(Self, CLIP_TEXT, NULL, NULL, NULL, 1);

         acFree(file);
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_DataChannel, ERR_CreateObject));
   }
   else DPrintF("@DataChannel:","[Clipboard] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  DeleteClip
Short: Retrieve the most recently clipped data as a list of files.

###DESCRIPTION###
<p>The DeleteClip method will clear all of the data that has been cached
against a particular datatype.  You may clear multiple datatypes by OR'ing
clip types in the Datatype parameter.  Refer to the AddClipboardFiles method
for a list of available clip types.  If you want to clear all content
from the clipboard, use the Clear action instead of this method.</p>

###INPUT###
LONG/Datatype: The datatype(s) that you want to delete (datatypes may be logically-or'd together).

###RESULTS###
ERR_Okay: The datatype was cleared successfully.
ERR_Args: The required arguments were not specified.
ERR_AccessMemory: The clipboard memory data was not accessible.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_DeleteClip(struct Clipboard *Self, struct mtDeleteClip *Args)
{
   struct ClipEntry *clips;
   WORD i;

   if ((!Args) OR (!Args->Datatype)) return(ObjectError(Self, MT_DeleteClip, ERR_Args));

   DPrintF("~DeleteClip()","[Clipboard:%d] Datatype: $%x", Self->Head.UniqueID, Args->Datatype);

   if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
      for (i=0; i < MAX_CLIPS; i++) {
         if (clips[i].Datatype & Args->Datatype) {
            FreeClip(Self, &clips[i]);
         }
      }
      ReleaseMemory(clips);
      StepBack();
      return(ERR_Okay);
   }
   else return(StepObjectError(Self, MT_DeleteClip, ERR_AccessMemory));
}

/*****************************************************************************
** Action: Free
*/

static ERROR CLIPBOARD_Free(struct Clipboard *Self, APTR Void)
{
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  GetClipboardFiles
Short: Retrieve the most recently clipped data as a list of files.

###DESCRIPTION###
<p>To retrieve items from a clipboard, you must use the GetClipboardFiles
method.  You must declare the types of data that you support (or NULL if
you recognise all datatypes) so that you are returned a recognisable
data format.</p>

<p>The most recent clip is always returned, so if you support text, audio
and image clips and the most recent clip is an audio file, any
text and image clips on the history buffer will be ignored by this routine.
If you want to scan all available clip items, set the Datatype parameter
to NULL and repeatedly call this method with incremented Index numbers until
the error code ERR_OutOfRange is returned.</p>

<p>On success this method will return a list of files (terminated with a
NULL entry) in the Files parameter.  Each file is a readable clipboard
entry - how you read it depends on the resulting Datatype - although you may use
the IdentifyFile() function to determine a class that
supports the data.  Note that the resulting Files array is a memory allocation that must be freed using
the FreeMemory() function.</p>

<p>If this method returns the CLF_DELETE flag in the Flags parameter,
you must delete the source files after successfully copying out the
data.  If you are not successful in your operation, do not proceed with the deletion or
the user will have lost the original data.  When cutting and pasting files
within the file system, simply using the file system's 'move file'
functionality may be useful for fast file transfer.</p>

###INPUT###
LONG/Datatype: The types of data that you recognise for retrieval are specified here (for example CLIP_TEXT, CLIP_AUDIO, CLIP_FILE...).  You may logical-OR multiple data types together.  This parameter will be updated to reflect the retrieved data type when the method returns.
LONG/Index:    If the Datatype parameter is NULL, this parameter may be set to the index of the clip item that you want to retrieve.
*STRING/Files: The resulting location(s) of the requested clip data are returned in this parameter; terminated with a NULL entry.  You are required to free the returned array with FreeMemory().
LONG/Flags:    Result flags are returned in this parameter.  If CLF_DELETE is set, you need to delete the files after use in order to support the 'cut' operation.

###RESULT###
ERR_Okay: A matching clip was found and returned.
ERR_Args: Arguments were not specified.
ERR_OutOfRange: The specified Index is out of the range of the available clip items.
ERR_NoData: No clip was available that matched the requested data type.

###END###

*****************************************************************************/

static ERROR CLIPBOARD_GetClipboardFiles(struct Clipboard *Self, struct mtGetClipboardFiles *Args)
{
   struct ClipEntry *clips;
   UBYTE buffer[100];
   STRING files, *list, str;
   ERROR error;
   WORD index, i;
   LONG len;

   if (!Args) return(ObjectError(Self, MT_GetClipboardFiles, ERR_Args));

   DPrintF("GetFiles()","[Clipboard:%d] Datatype: $%.8x", Self->Head.UniqueID, Args->Datatype);

   Args->Files = NULL;

   /*** Find the first clipboard entry to match what has been requested ***/

   if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
      if (!Args->Datatype) {
         /* Retrieve the most recent clip item, or the one indicated
         ** in the Index parameter.
         */
         if ((Args->Index < 0) OR (Args->Index >= MAX_CLIPS)) {
            ReleaseMemory(clips);
            return(ERR_OutOfRange);
         }

         index = Args->Index;
      }
      else {
         for (index=0; index < MAX_CLIPS; index++) {
            if (Args->Datatype & clips[index].Datatype) break;
         }
      }

      if (clips[index].TotalItems < 1) {
         DPrintF("@GetFiles:","No items are set for datatype $%x at clip index %d", clips[index].Datatype, index);
         ReleaseMemory(clips);
         return(ERR_NoData);
      }

      if (clips[index].Files) {
         if ((error = AccessMemory(clips[index].Files, MEM_READ, &files)) IS ERR_Okay) {
            for (len=0; files[len]; len++);
            len++;

            if (AllocMemory(((clips[index].TotalItems+1) * sizeof(STRING)) + len, MEM_DATA, &list, NULL) IS ERR_Okay) {
               CopyMemory(files, list + clips[index].TotalItems + 1, len);
            }
            else {
               ReleaseMemory(files);
               ReleaseMemory(clips);
               return(ObjectError(Self, MT_GetClipboardFiles, ERR_AllocMemory));
            }
            ReleaseMemory(files);
         }
         else {
            DPrintF("@GetFiles:","Failed to access file string #%d, error %d.", clips[index].Files, error);
            if (error IS ERR_MemoryDoesNotExist) clips[index].Files = NULL;
            ReleaseMemory(clips);
            return(ObjectError(Self, MT_GetClipboardFiles, ERR_AccessMemory));
         }
      }
      else {
         if (clips[index].Datatype IS CLIP_FILE) {
            DPrintF("@GetFiles:","File datatype detected, but no file list has been set.");
            ReleaseMemory(clips);
            return(ERR_Failed);
         }

         /*** Calculate the length of the file strings ***/

         StrFormat(buffer, sizeof(buffer), "temp:clipboard/%s%d.000", GetDatatype(clips[index].Datatype), clips[index].ID);
         len = ((StrLength(buffer) + 1) * clips[index].TotalItems) + 1;

         /* Allocate the list array with some room for the strings at
         ** the end and then fill it with data.
         */

         if (AllocMemory(((clips[index].TotalItems+1) * sizeof(STRING)) + len, MEM_DATA|MEM_CALLER, &list, NULL) IS ERR_Okay) {
            str = (STRING)(list + clips[index].TotalItems + 1);
            for (i=0; i < clips[index].TotalItems; i++) {
               if (i > 0) *str++ = ';';
               StrFormat(str, len, "temp:clipboard/%s%d.%.3d", GetDatatype(clips[index].Datatype), clips[index].ID, i);
               while (*str) str++;
            }
            *str = 0;
         }
         else {
            ReleaseMemory(clips);
            return(ObjectError(Self, MT_GetClipboardFiles, ERR_AllocMemory));
         }
      }

      /*** Setup the pointers in the string list ***/

      str = (STRING)(list + clips[index].TotalItems + 1);
      for (i=0; i < clips[index].TotalItems; i++) {
         list[i] = str;
         while ((*str) AND (*str != ';')) str++;
         if (*str IS ';') { *str = 0; str++; }
      }
      list[i] = NULL;

      /*** Results ***/

      Args->Datatype = clips[index].Datatype;
      Args->Files    = list;
      Args->Flags    = clips[index].Flags;

      ReleaseMemory(clips);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, MT_GetClipboardFiles, ERR_AccessMemory));
}

/*****************************************************************************

###ACTION###
Name:  GetUnlistedField
Short: Special field types are suppored via unlisted fields.

###DESCRIPTION###
<p>The following unlisted field types are supported by the Clipboard class:</p>

<p><i>File(Datatype, Index)</i> Where Datatype is a recognised data format (e.g.
TEXT) and Index is between 0 and the Items() field.  If you don't support
multiple clipped items, use an index of zero.  On success, this
field will return a file location that points to the clipped data.</p>

<p><i>Items(Datatype)</i> Returns the total number of items available for
the specified data type.</p>

###END###

*****************************************************************************/

static ERROR CLIPBOARD_GetUnlistedField(struct Clipboard *Self, struct acGetUnlistedField *Args)
{
   struct ClipEntry *clip, *clips;
   UBYTE datatype[20], index[20];
   STRING str, files;
   WORD i, j, total;
   LONG value;

   if (!Args) return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Args));

   if ((!Args->Field) OR (!Args->Buffer) OR (Args->Size < 1)) {
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Args));
   }

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_Failed));
   }

   Args->Buffer[0] = 0;

   if (StrCompare("File(", Args->Field, 0, NULL) IS ERR_Okay) {
      /*** Extract the datatype ***/

      for (j=0, i=6; (Args->Field[i]) AND (Args->Field[i] != ',') AND (Args->Field[i] != ')'); i++) datatype[j++] = Args->Field[i];
      datatype[j] = 0;

      /*** Convert the datatype string into its equivalent value ***/

      value = 0;
      for (i=0; glDatatypes[i].Name; i++) {
         if (StrCompare(datatype, glDatatypes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            value = glDatatypes[i].Value;
            break;
         }
      }

      if (Args->Field[i] IS ',') i++;
      while ((Args->Field[i]) AND (Args->Field[i] <= 0x20)) i++;

      for (j=0; (Args->Field[i]) AND (Args->Field[i] != ')'); i++) index[j++] = Args->Field[i];
      index[j] = 0;

      if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
         /*** Find the clip for the requested datatype ***/

         clip = NULL;
         for (i=0; i < MAX_CLIPS; i++) {
            if (clips[i].Datatype IS value) {
               clip = &clips[i];
               break;
            }
         }

         if (clip) {
            i = StrToInt(index);
            if ((i >= 0) AND (i < clip->TotalItems)) {
               if (clip->Files) {
                  if (AccessMemory(clip->Files, MEM_READ, &files) IS ERR_Okay) {
                     str = files;
                     while (i) {
                        while ((*str) AND (*str != ';')) str++;
                        if (*str IS ';') str++;
                        i--;
                     }

                     for (i=0; (str[i]) AND (str[i] != ';'); i++) Args->Buffer[i] = str[i];
                     Args->Buffer[i] = 0;

                     ReleaseMemory(files);
                  }
                  else {
                     ReleaseMemory(clips);
                     return(ObjectError(Self, ERH_GetUnlistedField, ERR_AccessMemory));
                  }
               }
               else StrFormat(Args->Buffer, Args->Size, "temp:clipboard/%s%d.%.3d", datatype, clip->ID, i);
            }
         }

         ReleaseMemory(clips);
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_GetUnlistedField, ERR_AccessMemory));
   }
   else if (StrCompare("Items(", Args->Field, 0, NULL) IS ERR_Okay) {
      /*** Extract the datatype ***/
      for (j=0, i=6; (Args->Field[i]) AND (Args->Field[i] != ')'); i++) datatype[j++] = Args->Field[i];
      datatype[j] = 0;

      /*** Convert the datatype string into its equivalent value ***/

      value = 0;
      for (i=0; glDatatypes[i].Name; i++) {
         if (StrCompare(datatype, glDatatypes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            value = glDatatypes[i].Value;
            break;
         }
      }

      /*** Calculate the total number of items available for this datatype ***/

      total = 0;
      if (value) {
         if (AccessMemory(RPM_Clipboard, MEM_READ, &clips) IS ERR_Okay) {
            for (i=0; i < MAX_CLIPS; i++) {
               if (clips[i].Datatype IS value) {
                  total = clips[i].TotalItems;
                  break;
               }
            }
            ReleaseMemory(clips);
         }
      }

      IntToStr(total, Args->Buffer, Args->Size);
      return(ERR_Okay);
   }
   else return(ERR_NoSupport);
}

/*****************************************************************************
** Clipboard: Init
*/

static ERROR CLIPBOARD_Init(struct Clipboard *Self, APTR Void)
{
   struct FileInfo *list;

   /*** Create a directory under temp: to store clipboard data ***/

   MakeDirectory("temp:clipboard", PERMIT_READ|PERMIT_WRITE);

   /* Scan the temp:clipboard directory for existing clips.  If they
   ** exist, add them to our clipboard object.  This allows the user to
   ** continue using clipboard data from the last time that the machine
   ** was powered on.
   */

   if (ReadDirectory("temp:clipboard/", NULL, &list) IS ERR_Okay) {

      FreeMemory(list);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:    PasteClipboardFiles
Synonym: PasteFiles
Short:   Performs the 'paste' operation on file items.

###DESCRIPTION###
<p>The PasteClipboardFiles method is used to implement file-paste operations
for the CLIP_FILE datatype in a convenient way for the developer.  The correct
handling for both copy and cut operations is implemented for you.</p>

<p>You are required to specify a destination path for the file items.  This
should be a directory (denoted by a trailing slash or colon character), although
you may specify a single target file if you wish.</p>

<p>Confirmation windows are automatically handled by this routine when duplicate
file names or errors are encountered during procesing.  Any windows will open on
whatever drawable you specify as the Target (this should be set to either the
Desktop object, or a screen that you have created).  If the Target is set to
NULL, the routine will discover the most appropriate space for opening new
windows. If the Target is set to -1, the routine will overwrite all files by
default and not inform the user if errors are encountered.</p>

###INPUT###
STRING/Destination: The destination path for the file items.  Must be identifiable as a directory location, otherwise the file(s) will be copied to a destination file.
LONG/Target: Set this parameter to a valid drawable and the paste operation will open interactive windows (such as progress bars and user confirmations) during the paste operation.  This is usually set to the "Desktop" object, which is also the default if this argument is NULL.  Use a value of -1 if you don't want to provide feedback to the user.

###RESULT###
ERR_Okay: The paste operation was performed successfully (this code is returned even if there are no files to be pasted).
ERR_Args: Invalid arguments were specified.
ERR_File: A non-specific file error occurred during the paste operation.

###END###

*****************************************************************************/

static STRING PasteConfirm = "STRING:<dml>\n\
  <if not exists=\"[PasteConfirm]\">\n\
    <dialog static name=\"PasteConfirm\" image=\"icons:items/warning(48)\"\n\
      buttons=\"yes;yesall;no;cancel\" title=\"Confirmation Required\"\n\
      flags=\"wait\" responseobject=\"[{scriptowner}]\" responsefield=\"response\"/>\n\
  </if>\n\
\n\
  <set object=\"[PasteConfirm]\" &string=\"You are about to overwrite this location - should it be replaced?\n\n{file}\"/>\n\
\n\
  <action call=\"show\" object=\"[PasteConfirm]\"/>\n\
</dml>";

static STRING PasteError = "STRING:<dml>\n\
  <dialog image=\"icons:items/warning(48)\" buttons=\"okay\" title=\"File Paste Failure\" flags=\"wait\"\n\
    string=\"An error occurred while pasting this file - the process has been terminated:\n\n{file}\"/>\n\
</dml>";

static ERROR CLIPBOARD_PasteClipboardFiles(struct Clipboard *Self, struct mtPasteClipboardFiles *Args)
{
   struct mtGetClipboardFiles get;
   OBJECTID dialogid;
   STRING *list;
   UBYTE dest[512];
   LONG type;
   ERROR error;
   WORD i, findex, dindex;

   if (!Args) return(ObjectError(Self, MT_PasteClipboardFiles, ERR_Args));
   if ((!Args->Destination) OR (!Args->Destination[0])) return(ObjectError(Self, MT_PasteClipboardFiles, ERR_MissingLocation));

   DPrintF("~PasteFiles()","[Clipboard:%d] Destination: %s, Target: %d", Self->Head.UniqueID, Args->Destination, Args->Target);

   Self->Response = RSP_YES;

   Self->ProgressTime = SystemTime();
   Self->ProgressTarget = Args->Target;
   Self->ProgressWindow = NULL;
   get.Datatype = CLIP_FILE;
   if (Action(MT_GetClipboardFiles, Self, &get) IS ERR_Okay) {
      /*** Scan the file list and move or copy each file to the destination directory ***/

      error = ERR_Okay;
      list = get.Files;
      for (i=0; list[i]; i++) {
         if (Self->Response != RSP_YESALL) Self->Response = RSP_YES; /* Reset Response to YES default */

         /*** Determine the index at which the source filename starts ***/

         for (findex=0; list[i][findex]; findex++);
         if ((list[i][findex-1] IS '/') OR (list[i][findex-1] IS ':') OR (list[i][findex-1] IS '\\')) findex--;
         while ((findex > 0) AND (list[i][findex-1] != '/') AND (list[i][findex-1] != ':') AND (list[i][findex-1] != '\\')) findex--;

         /*** Copy the source's filename to the end of the destination path ***/

         dindex = StrCopy(Args->Destination, dest, sizeof(dest)-1);
         if ((dest[dindex-1] IS '/') OR (dest[dindex-1] IS '\\') OR (dest[dindex-1] IS ':')) {
            StrCopy(list[i]+findex, dest+dindex, sizeof(dest)-dindex);
         }

         /*** Check if the destination already exists ***/

         if ((Self->Response != RSP_YESALL) AND (Args->Target != -1)) {
            if (StrCompare(list[i], dest, 0, STR_MATCHLENGTH) IS ERR_Okay) {
               if (get.Flags & CLF_DELETE) {
                  continue; /* Do nothing for move operations when the source and destination are identical */
               }
               else {
                  /* The source and destination strings are exactly identical.  In
                  ** this case, the destination will be "Copy of [Source File]"
                  */
                  if ((dest[dindex] >= 'A') AND (dest[dindex] <= 'Z')) dindex += StrCopy("Copy of ", dest+dindex, sizeof(dest)-dindex);
                  else dindex += StrCopy("copy of ", dest+dindex, sizeof(dest)-dindex);

                  dindex += StrCopy(list[i]+findex, dest+dindex, sizeof(dest)-dindex);
               }
            }

            /*** Request user confirmation ***/

            if ((AnalyseLocation(dest, &type) IS ERR_Okay) AND ((type IS LOC_FILE) OR (type IS LOC_DIRECTORY))) {
               Self->Response = RSP_CANCEL;
               RunScript(PasteConfirm, Args->Target, dest);
               if (!Self->ProgressWindow) Self->ProgressTime = SystemTime(); /* Reset the start time whenever a user dialog is presented */
            }
         }

         if (Self->Response IS RSP_CANCEL) break;

         if ((Self->Response != RSP_YES) AND (Self->Response != RSP_YESALL)) {
            DPrintF("PasteFiles:","Skipping file %s, response %d", list[i], Self->Response);
            continue;
         }

         /*** Copy/Move the file ***/

         while (dest[dindex]) dindex++;
         if ((dest[dindex-1] IS '/') OR (dest[dindex-1] IS '\\')) dest[dindex-1] = 0;

         FileFeedback(PasteFeedback, Self, NULL);

         if (get.Flags & CLF_DELETE) error = MoveLocation(list[i], dest);
         else error = CopyLocation(list[i], dest);

         /*** If an error occurred, tell the user and stop the pasting process ***/

         if (error IS ERR_Cancelled) {
            error = ERR_Okay;
            break;
         }
         else if (error) {
            if (Args->Target != -1) RunScript(PasteError, Args->Target, dest);
            error = ERR_File;
            break;
         }
      }

      DPrintF("PasteFiles:","%d files were pasted.", i);

      if (FastFindObject("PasteConfirm", ID_DIALOG, &dialogid, 1, NULL) IS ERR_Okay) {
         acFreeID(dialogid);
      }

      if (Self->ProgressWindow) {
         acFreeID(Self->ProgressWindow);
         Self->ProgressWindow = NULL;
      }

      FreeMemory(list);
      StepBack();
      return(error);
   }
   else {
      DPrintF("PasteFiles:","There are no files to paste from the clipboard.");
      StepBack();
      return(ERR_Okay);
   }
}

static LONG PasteFeedback(struct FileFeedback *Feedback)
{
   struct Clipboard *Self;
   OBJECTID tabfocusid;
   OBJECTPTR image, button, window;
   LONG leftmargin, topmargin, rightmargin, bottommargin, imagewidth, height, textheight;
   LONG buttonheight;
   UBYTE buffer[256];
   #define DLG_WIDTH   290
   #define DLG_HEIGHT  102

   Self = Feedback->User;

   if (Self->Response IS RSP_CANCEL) return(FFR_Abort);

   if ((!Self->ProgressWindow) AND (Self->ProgressTarget != -1) AND ((SystemTime() - Self->ProgressTime) > 500)) {
      /*** If the copy process exceeds 1 second, popup a progress window ***/

      if (CreateObject(ID_WINDOW, NF_CHILD, &window, &Self->ProgressWindow,
         FID_Name|TSTRING,        "winProgress",
         FID_Parent|TLONG,        Self->ProgressTarget,
         FID_Title|TSTRING,       "File Transfer Progress",
         FID_InsideWidth|TLONG,   DLG_WIDTH,
         FID_InsideHeight|TLONG,  DLG_HEIGHT,
         FID_MinWidth|TLONG,      DLG_WIDTH,
         FID_MinHeight|TLONG,     DLG_HEIGHT,
         FID_MaxWidth|TLONG,      DLG_WIDTH,
         FID_MaxHeight|TLONG,     DLG_HEIGHT,
         FID_Quit|TLONG,          FALSE,
         FID_Flags|TLONG,         WNF_MANUALCLOSE,
         FID_Icon|TSTRING,        "icons:tools/copy",
         TAGEND) IS ERR_Okay) {

         SubscribeActionTags(window, Self, MT_CloseWindow, TAGEND);

         GetFields(window, FID_LeftMargin|TLONG,   &leftmargin,
                           FID_TopMargin|TLONG,    &topmargin,
                           FID_BottomMargin|TLONG, &bottommargin,
                           FID_RightMargin|TLONG,  &rightmargin,
                           TAGEND);

         /*** Create tab-focus manager ***/

         CreateObject(ID_TABFOCUS, NULL, NULL, &tabfocusid,
            FID_Owner|TLONG, Self->ProgressWindow,
            TAGEND);

         /*** Image ***/

         if (CreateObject(ID_IMAGE, NULL, &image, NULL,
            FID_Owner|TLONG,      Self->ProgressWindow,
            FID_Location|TSTRING, "icons:tools/copy(48)",
            FID_XCoord|TLONG,     leftmargin,
            FID_YCoord|TLONG,     topmargin,
            TAGEND) IS ERR_Okay) {
            GetField(image, FID_ImageWidth, FT_LONG, &imagewidth);
            leftmargin += imagewidth + 10;
         }

         /*** Cancel button ***/

         if (CreateObject(ID_BUTTON, NULL, &button, NULL,
            FID_Owner|TLONG,    Self->ProgressWindow,
            FID_Text|TSTRING,   "Cancel",
            FID_YOffset|TLONG,  bottommargin,
            FID_XOffset|TLONG,  rightmargin,
            TAGEND) IS ERR_Okay) {

            CreateObject(ID_SET, NULL, NULL, NULL,
               FID_Owner|TLONG,        button->UniqueID,
               FID_Object|TLONG,       Self->Head.UniqueID,
               FID_Static|TLONG,       TRUE,
               FID_Response|TLONG|TUNLISTED, RSP_CANCEL,
               TAGEND);

            GetField(button, FID_Height, FT_LONG, &buttonheight);
            acShow(button);
         }

         /*** Progress text ***/

         if (CreateObject(ID_TEXT, NULL, &Self->ProgressText, NULL,
            FID_Owner|TLONG,    Self->ProgressWindow,
            FID_XCoord|TLONG,   leftmargin,
            FID_YCoord|TLONG,   topmargin,
            FID_XOffset|TLONG,  rightmargin,
            FID_Flags|TLONG,    TXF_WORDWRAP,
            FID_String|TSTRING, "Copying...",
            FID_Face|TSTRING|TTRANS,   "[gl_fonts.face]",
            FID_Colour|TSTRING|TTRANS, "[gl_fonts.window_colour]",
            TAGEND) IS ERR_Okay) {
            OBJECTPTR text;

            text = Self->ProgressText;

            if (GetField(text, FID_TextHeight, FT_LONG, &textheight) IS ERR_Okay) {
               if (textheight < 20) textheight = 20;

               height = topmargin + textheight + 30 + buttonheight + bottommargin;
               SetFields(window, FID_MinHeight|TLONG, height - bottommargin - topmargin,
                                 FID_MaxHeight|TLONG, height - bottommargin - topmargin,
                                 TAGEND);
               acResize(window, 0, height, 0);
            }
         }

         acShow(window);

         ReleaseObject(window);
      }
   }

   if (!Self->ProgressWindow) return(FFR_Okay);

   if ((Feedback->Position IS 0) OR (Feedback->Size > 32768)) {
      STRING str;
      WORD i;

      str = Feedback->Location;
      for (i=0; str[i]; i++);
      while ((i > 0) AND (str[i-1] != '/') AND (str[i-1] != '\\') AND (str[i-1] != ':')) i--;

      if (Feedback->Position IS 0) {
         if (Feedback->Size >= 1048576) StrFormat(buffer, sizeof(buffer), "Copying: %s\n\n%.2f MB", str+i, (DOUBLE)Feedback->Size / 1048576.0);
         else StrFormat(buffer, sizeof(buffer), "Copying: %s\n\n%.2f KB", str+i, (DOUBLE)Feedback->Size / 1024.0);
         SetField(Self->ProgressText, FID_String, FT_STRING, buffer);
         Self->ProgressTime = SystemTime();

         ProcessMessages(NULL, NULL);
      }
      else if (SystemTime() - Self->ProgressTime > 50) {
         if (Feedback->Size >= 1048576) StrFormat(buffer, sizeof(buffer), "Copying: %s\n\n%.2f / %.2f MB", str+i, (DOUBLE)Feedback->Position / 1048576.0, (DOUBLE)Feedback->Size / 1048576.0);
         else StrFormat(buffer, sizeof(buffer), "Copying: %s\n\n%.2f / %.2f KB", str+i, (DOUBLE)Feedback->Position / 1024.0, (DOUBLE)Feedback->Size / 1024.0);
         SetField(Self->ProgressText, FID_String, FT_STRING, buffer);
         Self->ProgressTime = SystemTime();

         ProcessMessages(NULL, NULL);
      }
   }

   return(FFR_Okay);
}

/*****************************************************************************
** Action: ReleaseObject()
*/

static ERROR CLIPBOARD_ReleaseObject(struct Clipboard *Self, APTR Void)
{
   Self->CrashProtect = 0; /* Turn off crash protection */
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: RunScript()
*/

static void RunScript(STRING Statement, OBJECTID Target, STRING File)
{
   OBJECTPTR script;

   if (Statement) {
      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
            FID_String|TSTRING,  Statement,
            FID_Target|TLONG,    Target,
            FID_File|TUNLISTED,  File,
            TAGEND) IS ERR_Okay) {
         acActivate(script);
         acFree(script);
      }
   }
}

/*****************************************************************************
** Internal: FreeClip()
*/

static void FreeClip(struct Clipboard *Self, struct ClipEntry *Clip)
{
   STRING datatype;
   UBYTE buffer[200];
   WORD i;

   if (Clip->TotalItems > 16384) Clip->TotalItems = 16384;

   if (Clip->Datatype != CLIP_FILE) {
      datatype = GetDatatype(Clip->Datatype);

      DPrintF("~FreeClip()","[Clipboard:%d] Deleting %d clip files for datatype %s / %d.", Self->Head.UniqueID, Clip->TotalItems, datatype, Clip->Datatype);

      /*** Delete cached clipboard files ***/
      for (i=0; i < Clip->TotalItems; i++) {
         StrFormat(buffer, sizeof(buffer), "temp:clipboard/%s%d.%.3d", datatype, Clip->ID, i);

         DeleteFile(buffer);
      }
   }
   else DPrintF("~FreeClip()","[Clipboard:%d] Datatype: File", Self->Head.UniqueID);

   if (Clip->Files) {
      FreeMemoryID(Clip->Files);
      Clip->Files = NULL;
   }

   ClearMemory(Clip, sizeof(struct ClipEntry));

   StepBack();
}

/*****************************************************************************
** Internal: AddClip()
*/

static void AddClip(struct Clipboard *Self, LONG Datatype, STRING Files, LONG Flags, CLASSID ClassID, LONG TotalItems)
{
   struct ClipEntry clip, tmp, *clips;
   STRING str;
   LONG len;
   WORD i;

   DPrintF("AddClip()","Datatype: $%x, Files: %s, Flags: $%x, Class: %d, Items: %d", Datatype, Files, Flags, ClassID, TotalItems);

   ClearMemory(&clip, sizeof(clip));

   if (!TotalItems) {
      DPrintF("AddClip:","TotalItems parameter not specified.");
      return;
   }

   if (AccessMemory(RPM_Clipboard, MEM_READWRITE, &clips) IS ERR_Okay) {
      if (Flags & CLF_EXTEND) {
         /*** Search for an existing clip that matches the requested datatype ***/
         for (i=0; i < MAX_CLIPS; i++) {
            if (clips[i].Datatype IS Datatype) {
               DPrintF("AddClip:","Found clip to extend.");

               /* We have found a matching datatype.  Start by moving the
               ** clip to the front of the queue.
               */

               CopyMemory(clips+i, &tmp, sizeof(struct ClipEntry));           /* !!! We need to shift the list down, not swap entries !!! */
               CopyMemory(clips, clips+i, sizeof(struct ClipEntry));
               CopyMemory(&tmp, clips, sizeof(struct ClipEntry));

               /*** Extend the existing clip with the new items/files ***/

               if (Files) {
                  if (clips->Files) {
                     if (AccessMemory(clips->Files, MEM_READWRITE, &str) IS ERR_Okay) {
                        len = StrLength(str);
                        if (ReallocMemory(str, len + StrLength(Files) + 2, &str, &clips->Files) IS ERR_Okay) {
                           str[len++] = ';';
                           len += StrCopy(Files, str + len, COPY_ALL);
                           str[len] = 0;
                           clips->TotalItems += TotalItems;
                        }
                        ReleaseMemory(str);
                     }
                  }
               }
               else clips->TotalItems += TotalItems;

               ReleaseMemory(clips);
               return;
            }
         }
      }

      /*** If a file string was specified, copy it to the clip entry ***/

      if (Files) {
         STRING str;
         if (AllocMemory(StrLength(Files) + 1, MEM_STRING|MEM_PUBLIC|MEM_UNTRACKED, &str, &clip.Files) IS ERR_Okay) {
            StrCopy(Files, str, COPY_ALL);
            ReleaseMemory(str);
         }
         else {
            ReleaseMemory(clips);
            return;
         }
      }

      /*** Set the clip details ***/

      clip.Datatype   = Datatype;
      clip.Flags      = Flags & (CLF_DELETE);
      clip.ClassID    = ClassID;
      clip.TotalItems = TotalItems;
      clip.ID         = ++Self->Counter;

      /*** Remove any existing clips that match this data type number ***/

      for (i=0; i < MAX_CLIPS; i++) {
         if (clips[i].Datatype IS Datatype) FreeClip(Self, &clips[i]);
      }

      /* If an entry has to be deleted due to a
      ** full historical buffer, kill it off.
      */

      if (clips[MAX_CLIPS-1].Datatype) {
         FreeClip(Self, &clips[MAX_CLIPS-1]);
      }

      /*** Insert the new clip entry at start of the history buffer ***/

      CopyMemory(clips, clips+1, MAX_CLIPS-1);
      CopyMemory(&clip, clips, sizeof(clip));

      ReleaseMemory(clips);
   }
}
