/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Keypress
Version:   1.0
Status:    Stable
ID:        ID_KEYPRESS
Category:  I/O
Date:      February 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Short:     Provides management for specific keys on the keyboard.
Keywords:  keyboard, key, keypress, down, up, release, input, output

###DESCRIPTION###
<p>The Keypress class is commonly used to respond to key-presses in the user
interface.  Special key codes such as the function keys, key combinations
like CTRL-Z, and miscellaneous keys such as PrintScreen can all be detected
by this class.  By default, key-presses are responded to when certain objects
(such as a window area) have the focus.  However, you can also support
key-presses globally, which is appropriate for keys such as PrintScreen.</p>

<p>The following example illustrates the use of a Keypress object that
is used to take screenshots whenever the PrintScreen button is pressed:</p>

<pre>
    &lt;keypress key="prtscr" flags="global"&gt;
       &lt;script static src="graphics:screenshot.dml"/&gt;
    &lt;/keypress&gt;
</pre>

<p>For a list of supported key-presses, refer to the documentation on the
<a href="#Key">Key</a> field.  Key qualifiers such as shift, control and alt
are also supported under the <a href="#Qualifier">Qualifier</a> field.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/input/all.h>
#include <pandora/graphics/render.h>
#include <pandora/input/keymaptable.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_KEYPRESS 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_KEYPRESS, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "February 2003",
   "Keypress"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR KeypressClass = 0;

/*****************************************************************************
** Class definition
*/

#define LEN_KEYCODE 30

struct Keypress {
   OBJECT_HEADER
   OBJECTID KeyboardID;
   LONG     Flags;
   OBJECTID FocusID;
   LONG     QualifierFlags;    /* Qualifier flags to check for */

   PRIVATE_FIELDS
   UBYTE Key[LEN_KEYCODE];
   LONG  KeyCode;              /* The key code, as translated from the key field */
   WORD  KeyboardChannel;      /* Set to TRUE if we have an open channel */
};

static ERROR GET_Key(struct Keypress *, STRING *);

static ERROR SET_Key(struct Keypress *, STRING);

#define KF_GLOBAL 0x00000001
#define KF_REPEAT 0x00000002

static struct FieldDef KeyFlags[] = {
   { "GLOBAL", KF_GLOBAL },
   { "REPEAT", KF_REPEAT },
   { NULL, NULL }
};

static struct FieldDef QualifierFlags[] = {
   { "LSHIFT",   KQ_LSHIFT   },
   { "RSHIFT",   KQ_RSHIFT   },
   { "CAPSLOCK", KQ_CAPSLOCK },
   { "LCTRL",    KQ_LCONTROL },
   { "RCTRL",    KQ_RCONTROL },
   { "LALT",     KQ_LALT     },
   { "RALT",     KQ_RALT     },
   { "LCOMMAND", KQ_LCOMMAND },
   { "RCOMMAND", KQ_RCOMMAND },
   { "NUMPAD",   KQ_NUMPAD   },
   /*** Pairs ***/
   { "SHIFT",    KQ_SHIFT   },
   { "COMMAND",  KQ_COMMAND },
   { "ALT",      KQ_ALT     },
   { "CTRL",     KQ_CONTROL },
   { "CONTROL",  KQ_CONTROL },
   { NULL, NULL }
};

static struct FieldArray KeypressFields[] = {
   { "Keyboard",  0, FDF_OBJECTID|FDF_RW,  0,               NULL, NULL },
   { "Flags",     0, FDF_LONGFLAGS|FDF_RW, (LONG)&KeyFlags, NULL, NULL },
   { "Focus",     0, FDF_OBJECTID|FDF_RI,  0,               NULL, NULL },
   { "Qualifier", 0, FDF_LONGFLAGS|FDF_RW, (LONG)&QualifierFlags, NULL, NULL },
   /*** Virtual fields ***/
   { "Key",       0, FDF_STRING|FDF_RW,    0, GET_Key, SET_Key },
   END_FIELD
};

/*****************************************************************************
** Class actions
*/

static ERROR KEYPRESS_ActionNotify(struct Keypress *, struct acActionNotify *);
static ERROR KEYPRESS_Activate(struct Keypress *, APTR);
static ERROR KEYPRESS_DataChannel(struct Keypress *, struct acDataChannel *);
static ERROR KEYPRESS_Init(struct Keypress *, APTR);
static ERROR KEYPRESS_Free(struct Keypress *, APTR);

static struct ActionArray KeypressActions[] = {
   { AC_ActionNotify,  KEYPRESS_ActionNotify },
   { AC_Activate,      KEYPRESS_Activate },
   { AC_DataChannel,   KEYPRESS_DataChannel },
   { AC_Free,          KEYPRESS_Free },
   { AC_Init,          KEYPRESS_Init },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&KeypressClass, NULL,
      FID_BaseClassID|TLONG, ID_KEYPRESS,
      FID_Version|TFLOAT,    VER_KEYPRESS,
      FID_Name|TSTRING,      "Keypress",
      FID_Category|TLONG,    CCF_IO,
      FID_Actions|TPTR,      KeypressActions,
      FID_Fields|TPTR,       KeypressFields,
      FID_Size|TLONG,        sizeof(struct Keypress),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (KeypressClass) { acFree(KeypressClass); KeypressClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Keypress: ActionNotify
*/

static ERROR KEYPRESS_ActionNotify(struct Keypress *Self, struct acActionNotify *Args)
{
   OBJECTPTR Keyboard;

   if (!Args) return(ERR_Args);
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if (Args->ActionID IS AC_Focus) {
      if (AccessObject(Self->KeyboardID, 5000, &Keyboard) IS ERR_Okay) {
         if (SubscribeChannel(Keyboard, Self) IS ERR_Okay) Self->KeyboardChannel = TRUE;
         ReleaseObject(Keyboard);
      }
   }
   else if (Args->ActionID IS AC_LostFocus) {
      if (AccessObject(Self->KeyboardID, 5000, &Keyboard) IS ERR_Okay) {
         UnsubscribeChannel(Keyboard, Self->Head.UniqueID);
         ReleaseObject(Keyboard);
      }
      Self->KeyboardChannel = FALSE;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Keypress: Activate
*/

static ERROR KEYPRESS_Activate(struct Keypress *Self, APTR Void)
{
   struct ChildEntry list[20];
   LONG count, j;

   count = ARRAYSIZE(list);
   if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
      for (j=0; j < count; j++) acActivateID(list[j].ObjectID);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  DataChannel
Short: This action can be used to send fake keyboard input to a Keypress object.

###DESCRIPTION###
<p>Fake keypresses can be sent to a Keypress object with the DATA_KEYPRESS
data type, as if the user was typing at the keyboard personally.  The
keypresses will be interpreted no differently to 'real' user keypresses.</p>

<p>It is extremely important that the key qualifiers are set correctly and
that each key press is matched with a key release.  If the data is not
defined correctly, it may be misinterpreted.</p>

###END###

*****************************************************************************/

static ERROR KEYPRESS_DataChannel(struct Keypress *Self, struct acDataChannel *Args)
{
   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version IS 1)) {
      struct dcKeyEntry *Key;
      LONG i;

      if ((Key = (struct dcKeyEntry *)Args->Buffer)) {
         for (i=0; i < Args->TotalEntries; i++) {
            //DPrintF("Keypress:","Flags: $%.8x : Char: $%.8x     ($%.8x, $%.8x)", Key[i].Flags, Key[i].Value, Self->QualifierFlags, Self->KeyCode);

            if ((Key[i].Flags & KQ_REPEAT) AND (!(Self->Flags & KF_REPEAT))) continue;

            if ((Key[i].Flags & KQ_PRESSED) AND (Key[i].Value IS Self->KeyCode)) {
               if (Self->QualifierFlags) {
                  if (Self->QualifierFlags & (KQ_LCONTROL|KQ_RCONTROL)) {
                     if (!((Key[i].Flags & (KQ_LCONTROL|KQ_RCONTROL)) & Self->QualifierFlags)) break;
                  }

                  if (Self->QualifierFlags & (KQ_LSHIFT|KQ_RSHIFT)) {
                     if (!((Key[i].Flags & (KQ_LSHIFT|KQ_RSHIFT)) & Self->QualifierFlags)) break;
                  }

                  if (Self->QualifierFlags & (KQ_LALT|KQ_RALT)) {
                     if (!((Key[i].Flags & (KQ_LALT|KQ_RALT)) & Self->QualifierFlags)) break;
                  }

                  if (Self->QualifierFlags & (KQ_LCOMMAND|KQ_RCOMMAND)) {
                     if (!((Key[i].Flags & (KQ_LCOMMAND|KQ_RCOMMAND)) & Self->QualifierFlags)) break;
                  }
               }

               acActivate(Self);
            }
         }
      }
   }
   else if (Args->DataType IS DATA_TEXT) {

   }
   else DPrintF("!DataChannel:","[Keypress] Unrecognised data type %d, version %d.", Args->DataType, Args->Version);

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Free
*/

static ERROR KEYPRESS_Free(struct Keypress *Self, APTR Void)
{
   OBJECTPTR object;

   /*** Close the keyboard channel ***/

   if ((Self->KeyboardID) AND (Self->KeyboardChannel IS TRUE)) {
      if (AccessObject(Self->KeyboardID, 3000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
   }

   if (Self->FocusID) {
      if (AccessObject(Self->FocusID, 2000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Action: Init
*/

static ERROR KEYPRESS_Init(struct Keypress *Self, APTR Void)
{
   LONG ObjectCount;
   OBJECTPTR Keyboard, object;
   OBJECTID OwnerID;

   if (!Self->KeyCode) return(ObjectError(Self, ERH_Init, ERR_FieldNotSet));

   if (!Self->KeyboardID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->KeyboardID, 1, &ObjectCount) != ERR_Okay) {
         DPrintF("@Init:","[Keypress] Could not find any SystemKeyboard objects.");
         return(ERR_Failed);
      }
   }

   /* Open a channel to the keyboard object if we are global.  Otherwise,
   ** subscribe to the Focus and LostFocus actions of the object that
   ** the user has requested us to monitor.
   */

   if (Self->Flags & KF_GLOBAL) {
      if (AccessObject(Self->KeyboardID, 5000, &Keyboard) IS ERR_Okay) {
         if (SubscribeChannel(Keyboard, Self) IS ERR_Okay) Self->KeyboardChannel = TRUE;
         ReleaseObject(Keyboard);
      }
   }
   else {
      if (!Self->FocusID) {
         OwnerID = GetOwner(Self);
         while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
            OwnerID = GetOwnerID(OwnerID);
         }
         if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
         else Self->FocusID = OwnerID;
      }

      if (AccessObject(Self->FocusID, 5000, &object) IS ERR_Okay) {
         SubscribeActionTags(object, Self, AC_Focus, AC_LostFocus, TAGEND);

         /* Test the focus object to see if it already has the focus and
         ** if so, activate keyboard monitoring.
         */

         if (object->ClassID IS ID_RENDER) {
            if (((struct Render *)object)->Flags & RNF_HASFOCUS) {
               if (AccessObject(Self->KeyboardID, 5000, &Keyboard) IS ERR_Okay) {
                  if (SubscribeChannel(Keyboard, Self) IS ERR_Okay) Self->KeyboardChannel = TRUE;
                  ReleaseObject(Keyboard);
               }
            }
         }

         ReleaseObject(object);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags.
Type:   LONG
Prefix: KF
Status: Read/Write

###DESCRIPTION###
<p>Special flags affecting the functionality of keypress objects are listed
in the following table:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">GLOBAL</td><td>If the global flag is set then the keypress object will monitor its associated key on a constant basis, rather than monitoring it at times when it has the focus.</td></tr>
<tr><td width="10%">REPEAT</td><td>Set the repeat flag if you would like the keypress object to respond to repeat keypresses (which occur when the user holds down a key for an extended period of time).</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Key
Short:  The key that is to be monitored.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>This field must refer to the key that the Keypress object will monitor and
respond to.  If you want to refer to a printable character such as that
represented by an alpha or numeric key, simply specify that character, e.g.
'a', '1', '='.  If you want to refer to a special key type such as a function
key, you need to choose the correct key from the following table:</p>

<table border="0" cellspacing="2" cellpadding="2" width="90%" align="center"><tr>
<td valign="top">
<p><b><u><font face="Arial, Helvetica, sans-serif">Function&nbsp;Keys</font></u></b><br>
f1<br>f2<br>f3<br>f4<br>f5<br>f6<br>f7<br>f8<br>f9<br>f10<br>f11<br>f12<br>f13<br>f14<br>f15<br>f16<br>f17<br>f18<br>f19<br>f20</p>
</td>
<td valign="top">
<p><b><u><font face="Arial, Helvetica, sans-serif">Cursor&nbsp;Keys</font></u></b><br>
down<br>left<br>right<br>up</p>
<p><b><u><font face="Arial, Helvetica, sans-serif">Control&nbsp;Keys</font></u></b><br>
lalt<br>ralt<br>lcommand<br>rcommand<br>lctrl<br>rctrl<br>lshift<br>rshift</p>
</td>
<td valign="top">
<p><b><u><font face="Arial, Helvetica, sans-serif">Orientation&nbsp;Keys</font></u></b><br>
end<br>home<br>pagedown<br>pageup</p>
<p><b><u><font face="Arial, Helvetica, sans-serif">Text&nbsp;Keys</font></u></b><br>
backspace<br>capslock<br>delete<br>enter<br>escape<br>insert<br>tab</p>
</td>
<td valign="top">
<p><b><u><font face="Arial, Helvetica, sans-serif">Special&nbsp;Keys</font></u></b><br>
break<br>cancel<br>clear<br>execute<br>find<br>help<br>menu<br>numlock<br>pause<br>power<br>prtscr<br>redo<br>scroll<br>select<br>sleep<br>undo<br>wake</p>
</td></tr>
</table>

<p>You can also set key qualifiers to handle combinations such as CTRL-C and
SHIFT-ENTER.  Refer to the <a href="#Qualifier">Qualifier</a> field for
details.</p>

###SEE ALSO###
Field: Qualifier

###END###

*****************************************************************************/

static ERROR GET_Key(struct Keypress *Self, STRING *Value)
{
   if (!Self->Key[0]) *Value = NULL;
   else *Value = Self->Key;
   return(ERR_Okay);
}

static ERROR SET_Key(struct Keypress *Self, STRING Value)
{
   LONG i, j;

   Self->Key[0] = 0;

   if ((Value) AND (*Value)) {
      for (i=0; i < K_LIST_END; i++) {
         if (StrCompare(Value, glKeymapTable[i], NULL, STR_MATCHLENGTH) IS ERR_Okay) {
            Self->KeyCode = i;
            for (j=0; (glKeymapTable[i][j]) AND (j < LEN_KEYCODE-1); j++) Self->Key[j] = glKeymapTable[i][j];
            Self->Key[j] = 0;
            return(ERR_Okay);
         }
      }
      DPrintF("!Keypress:","I do not recognise key \"%s\".", Value);
      return(ERR_Search);
   }
   else return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   KeyboardID
Short:  Must refer to a Keyboard object in order to receive key presses.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>By default a new Keypress object will read its input from the default
Keyboard object 'SystemKeyboard', which normally represents the keyboard
hooked into the PC.  If necessary you can set this field to explicitly
declare the Keyboard object that the keys should be read from.  Refer to the
<a href="keyboard.html">Keyboard</a> class for details on setting up Keyboard
objects.</p>

###SEE ALSO###
Class: Keyboard

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Focus
Short:  Determines the local object that will be monitored for key presses.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>To specify the object that you want to monitor for key presses, set this
field.  The Keypress object will watch the target object for key presses.
Incoming key presses will be compared to the Key and Qualifier fields to
check for a match.  Key presses are only translated when the target object
has the focus - at all other times, key presses are ignored.</p>

<p>You do not need to set the Focus field if you are checking for global
key presses, or if you want to monitor the drawable container of a Keypress
object.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Qualifier
Short:  Set this field to analyse key qualifiers such as shift and control.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>The Qualifier field can be set in conjunction with the Key field in order
to watch for key combinations.  The following table illustrates the qualifier
keys that are supported:</p>

<table border="0" cellspacing="2" cellpadding="2" width="90%" align="center"><tr>
<td valign="top">
<p><b><u>Paired&nbsp;Keys</u></b><br>
lalt<br>ralt<br>lctrl<br>rctrl<br>lcommand<br>rcommand<br>lshift<br>rshift</p>
</td>
<td valign="top">
<p><b><u>General&nbsp;Key&nbsp;References</u></b><br>
alt<br>capslock<br>command<br>ctrl<br>numpad<br>shift</p>
</td>
</table>

###SEE ALSO###
Field: Key

###END###

*****************************************************************************/
